const inlineNodesRegexp = new RegExp([
    '(?<quote>^(?<quoteTag>> )(?<quoteContent>\\S+))',
    '(?<tag>#(?<tagContent>\\S+))',
    '(?<bold>\\[\\[(?<boldContent>.+)\\]\\])',
    '(?<deco>\\[(?<decoKeys>[*/_-]+?) (?<decoContent>.+?)\\])',
    '(?<decoFormula>\\[(?<decoFormulaContent>\\$ (?<formula>.+?))\\])',
    '(?<icon>\\[(\\/(?<iconProject>\\S+?)\\/)?(?<iconContent>(?<iconPage>\\S+)?\\.icon)\\])',
    '(?<code>`(?<codeContent>[^`]*?)`)',
    '(?<gyazo>\\[(?<gyazoContent>https?://i\\.gyazo\\.com/\\S+)\\])',
    '(?<urlLinkB>\\[(?<urlLinkBContent>(?<!https?:)(?<urlLinkBTitle>\\S+)\\s+(?<urlLinkBLink>https?://\\S+))\\])',
    '(?<urlLinkA>\\[(?<urlLinkAContent>(?<urlLinkALink>https?://\\S+)(?:\\s+(?<urlLinkATitle>\\S+))?)\\])',
    '(?<link>\\[(?<linkContent>(?:/(?<linkProject>\\S|[^/]+)/)?(?<linkPage>\\S+?))\\])',
    '(?<url>https?://\\S+)',
    '(?<text>.+?(?=\\[.+?\\]|#\\S+|`[^`]*?`|https?://\\S+)|.+(?!\\[.+?\\]|#\\S+|`[^`]*?`|https?://\\S+))',
    '(?<throwError>.+)',
].join('|'), 'gu');
const parseInlineNodes = (text) => {
    if (text === '') {
        return '';
    }
    const nodes = [...text.matchAll(inlineNodesRegexp)]
        .map((match) => (match.groups ?? {}))
        .map((match) => {
        switch (true) {
            case typeof match.text === 'string': {
                return match.text;
            }
            case typeof match.quote === 'string': {
                return {
                    type: 'quote',
                    unit: {
                        content: match.quoteContent,
                        tag: match.quoteTag,
                        whole: match.quote,
                    },
                };
            }
            case typeof match.tag === 'string': {
                return {
                    type: 'hashTag',
                    unit: {
                        content: match.tagContent,
                        page: match.tagContent,
                        tag: '#',
                        whole: match.tag,
                    },
                };
            }
            case typeof match.bold === 'string': {
                return {
                    type: 'strong',
                    unit: {
                        content: match.boldContent,
                        whole: match.bold,
                    },
                };
            }
            case typeof match.deco === 'string': {
                return {
                    type: 'deco',
                    unit: {
                        content: match.decoContent,
                        deco: match.decoKeys,
                        italic: match.decoKeys.includes('/'),
                        strike: match.decoKeys.includes('-'),
                        underline: match.decoKeys.includes('_'),
                        strong: [...match.decoKeys].filter((char) => char === '*').length,
                        whole: match.deco,
                    },
                };
            }
            case typeof match.decoFormula === 'string': {
                return {
                    type: 'deco-formula',
                    unit: {
                        content: match.decoFormulaContent,
                        formula: match.formula,
                        whole: match.decoFormula,
                    },
                };
            }
            case typeof match.icon === 'string': {
                return {
                    type: 'icon',
                    unit: {
                        content: match.iconContent,
                        project: match.iconProject,
                        page: match.iconPage,
                        size: 1,
                        whole: match.icon,
                    },
                };
            }
            case typeof match.code === 'string': {
                return {
                    type: 'code',
                    unit: {
                        content: match.codeContent,
                        whole: match.code,
                    },
                };
            }
            case typeof match.link === 'string': {
                return {
                    type: 'link',
                    unit: {
                        whole: match.link,
                        content: match.linkContent,
                        page: match.linkPage,
                        project: match.linkProject,
                    },
                };
            }
            case typeof match.urlLinkA === 'string': {
                return {
                    type: 'urlLink',
                    unit: {
                        whole: match.urlLinkA,
                        content: match.urlLinkAContent,
                        link: match.urlLinkALink,
                        title: match.urlLinkATitle,
                    },
                };
            }
            case typeof match.urlLinkB === 'string': {
                return {
                    type: 'urlLink',
                    unit: {
                        whole: match.urlLinkB,
                        content: match.urlLinkBContent,
                        link: match.urlLinkBLink,
                        title: match.urlLinkBTitle,
                    },
                };
            }
            case typeof match.gyazo === 'string': {
                return {
                    type: 'gyazo',
                    unit: {
                        whole: match.gyazo,
                        content: match.gyazoContent,
                    },
                };
            }
            case typeof match.url === 'string': {
                return {
                    type: 'url',
                    unit: {
                        whole: match.url,
                        content: match.url,
                    },
                };
            }
            default: {
                throw new Error('unreachable');
            }
        }
    });
    return nodes.length === 1 ? nodes[0] : nodes;
};

export { parseInlineNodes };
//# sourceMappingURL=inline-nodes.js.map
